@@ -15,6 +15,7 @@ limitations under the License.
 """
 
 # Custom utils import
+from google.cloud.sql.connector.rate_limiter import AsyncRateLimiter
 from google.cloud.sql.connector.refresh_utils import _get_ephemeral, _get_metadata
 from google.cloud.sql.connector.utils import write_to_file
 from google.cloud.sql.connector.version import __version__ as version
@@ -26,7 +27,7 @@ import concurrent
 import datetime
 from enum import Enum
 import google.auth
-from google.auth.credentials import Credentials
+from google.auth.credentials import Credentials, with_scopes_if_required
 import google.auth.transport.requests
 import OpenSSL
 import platform
@@ -116,6 +117,15 @@ class PlatformNotSupportedError(Exception):
         super(PlatformNotSupportedError, self).__init__(self, *args)
 
 
+class CredentialsTypeError(Exception):
+    """
+    Raised when credentials parameter is not proper type.
+    """
+
+    def __init__(self, *args: Any) -> None:
+        super(CredentialsTypeError, self).__init__(self, *args)
+
+
 class InstanceMetadata:
     ip_addrs: Dict[str, Any]
     context: ssl.SSLContext
@@ -176,6 +186,11 @@ class InstanceConnectionManager:
         The user agent string to append to SQLAdmin API requests
     :type user_agent_string: str
 
+    :type credentials: google.auth.credentials.Credentials
+    :param credentials
+        Credentials object used to authenticate connections to Cloud SQL server.
+        If not specified, Application Default Credentials are used.
+
     :param enable_iam_auth
         Enables IAM based authentication for Postgres instances.
     :type enable_iam_auth: bool
@@ -218,8 +233,9 @@ class InstanceConnectionManager:
     _project: str
     _region: str
 
-    _current: asyncio.Task
-    _next: asyncio.Task
+    _refresh_in_progress: asyncio.locks.Event
+    _current: asyncio.Task  # task wraps coroutine that returns InstanceMetadata
+    _next: asyncio.Task  # task wraps coroutine that returns another task
 
     def __init__(
         self,
@@ -227,6 +243,7 @@ class InstanceConnectionManager:
         driver_name: str,
         keys: concurrent.futures.Future,
         loop: asyncio.AbstractEventLoop,
+        credentials: Optional[Credentials] = None,
         enable_iam_auth: bool = False,
     ) -> None:
         # Validate connection string
@@ -248,10 +265,22 @@ class InstanceConnectionManager:
         self._user_agent_string = f"{APPLICATION_NAME}/{version}+{driver_name}"
         self._loop = loop
         self._keys = asyncio.wrap_future(keys, loop=self._loop)
-        self._auth_init()
+        # validate credentials type
+        if not isinstance(credentials, Credentials) and credentials is not None:
+            raise CredentialsTypeError(
+                "Arg credentials must be type 'google.auth.credentials.Credentials' "
+                "or None (to use Application Default Credentials)"
+            )
+
+        self._auth_init(credentials)
+
+        self._refresh_rate_limiter = AsyncRateLimiter(
+            max_capacity=2, rate=1 / 30, loop=self._loop
+        )
 
         async def _set_instance_data() -> None:
             logger.debug("Updating instance data")
+            self._refresh_in_progress = asyncio.locks.Event()
             self._current = self._loop.create_task(self._get_instance_data())
             self._next = self._loop.create_task(self._schedule_refresh())
 
@@ -336,20 +365,57 @@ class InstanceConnectionManager:
             self._enable_iam_auth,
         )
 
-    def _auth_init(self) -> None:
+    def _auth_init(self, credentials: Optional[Credentials]) -> None:
         """Creates and assigns a Google Python API service object for
         Google Cloud SQL Admin API.
-        """
 
-        credentials, project = google.auth.default(
-            scopes=[
-                "https://www.googleapis.com/auth/sqlservice.admin",
-                "https://www.googleapis.com/auth/cloud-platform",
-            ]
-        )
+        :type credentials: google.auth.credentials.Credentials
+        :param credentials
+            Credentials object used to authenticate connections to Cloud SQL server.
+            If not specified, Application Default Credentials are used.
+        """
+        scopes = [
+            "https://www.googleapis.com/auth/sqlservice.admin",
+            "https://www.googleapis.com/auth/cloud-platform",
+        ]
+        # if Credentials object is passed in, use for authentication
+        if isinstance(credentials, Credentials):
+            credentials = with_scopes_if_required(credentials, scopes=scopes)
+        # otherwise use application default credentials
+        else:
+            credentials, project = google.auth.default(scopes=scopes)
 
         self._credentials = credentials
 
+    async def _force_refresh(self) -> bool:
+        if self._refresh_in_progress.is_set():
+            # if a new refresh is already in progress, then block on the result
+            self._current = await self._next
+            return True
+        try:
+            self._next.cancel()
+            # schedule a refresh immediately with no delay
+            self._next = self._loop.create_task(self._schedule_refresh(0))
+            self._current = await self._next
+            return True
+        except Exception as e:
+            # if anything else goes wrong, log the error and return false
+            logger.exception("Error occurred during force refresh attempt", exc_info=e)
+            return False
+
+    def force_refresh(self, timeout: Optional[int] = None) -> bool:
+        """
+        Forces a new refresh attempt and returns a boolean value that indicates
+        whether the attempt was successful.
+
+        :type timeout: Optional[int]
+        :param timeout: Amount of time to wait for the attempted force refresh
+        to complete before throwing a timeout error.
+        """
+        return asyncio.run_coroutine_threadsafe(
+            self._force_refresh(), self._loop
+        ).result(timeout=timeout)
+
     async def seconds_until_refresh(self) -> int:
         expiration = (await self._current).expiration
 
@@ -378,7 +444,8 @@ class InstanceConnectionManager:
         :rtype: concurrent.future.Futures
         :returns: A future representing the creation of an SSLcontext.
         """
-
+        self._refresh_in_progress.set()
+        await self._refresh_rate_limiter.acquire()
         logger.debug("Entered _perform_refresh")
 
         refresh_task = self._loop.create_task(self._get_instance_data())
@@ -387,7 +454,8 @@ class InstanceConnectionManager:
             await refresh_task
         except Exception as e:
             logger.exception(
-                "An error occurred while performing refresh. Retrying in 60s.",
+                "An error occurred while performing refresh."
+                "Scheduling another refresh attempt immediately",
                 exc_info=e,
             )
             instance_data = None
@@ -401,14 +469,14 @@ class InstanceConnectionManager:
                 or instance_data.expiration < datetime.datetime.now()
             ):
                 self._current = refresh_task
-                # TODO: Implement force refresh method and a rate-limiter for perform_refresh
-                # Retry by scheduling a refresh 60s from now.
-                self._next = self._loop.create_task(self._schedule_refresh(60))
+                self._next = self._loop.create_task(self._perform_refresh())
 
         else:
             self._current = refresh_task
             # Ephemeral certificate expires in 1 hour, so we schedule a refresh to happen in 55 minutes.
             self._next = self._loop.create_task(self._schedule_refresh())
+        finally:
+            self._refresh_in_progress.clear()
 
         return refresh_task
 
@@ -419,17 +487,15 @@ class InstanceConnectionManager:
         :rtype: asyncio.Task
         :returns: A Task representing _get_instance_data.
         """
-        logger.debug("Entering sleep")
 
         if delay is None:
             delay = await self.seconds_until_refresh()
-
         try:
+            logger.debug("Entering sleep")
             await asyncio.sleep(delay)
         except asyncio.CancelledError as e:
             logger.debug("Schedule refresh task cancelled.")
             raise e
-
         return await self._perform_refresh()
 
     def connect(
